@@ -30,18 +30,22 @@ int within_line_edit = FALSE;	     /* TRUE while user is editing input */
 int always_readline = FALSE;	     /* -a option */
 char *password_prompt_search_string  /* argument of -a option (cf manpage) */
    = NULL;
+char *history_format = NULL;         /* format for history entries */
 int complete_filenames = FALSE;	     /* -c option */
 int child_pid = 0;		     /* pid of child (client) */
 int i_am_child = FALSE;		     /* after forking, child will set this to TRUE */
 int nowarn = FALSE;		     /* argument of -n option (suppress warnings) */
 int debug = 0;			     /* debugging mask (0 = no debugging) */
 int ignore_queued_input = FALSE;     /* read and then ignore all characters in input queue until it is empty (i.e. read would block) */
+int history_duplicate_avoidance_policy =
+   ELIMINATE_SUCCESIVE_DOUBLES;      /* whether and how to avoid duplicate history entries */
+int one_shot_rlwrap = FALSE;
 
 /* private vars */
 
 static char *history_filename = NULL;
-static int histsize = 300;
-static int write_histfile = TRUE;
+static int  histsize = 300;
+static int  write_histfile = TRUE;
 static char *completion_filename, *default_completion_filename;
 static char *full_program_name;
 static void init_rlwrap(void);
@@ -49,48 +53,26 @@ static void fork_child(char *command_name, char **argv);
 static char *read_options_and_command_name(int argc, char **argv);
 static void main_loop(void);
 static void flush_output_queue(void);
-static int output_queue_is_nonempty(void);
-
-#ifdef GETOPT_GROKS_OPTIONAL_ARGS
-static char optstring[] = "+:a::b:cC:d::f:hH:il:nm::rs:v";	/* +: is not really documented. configure checks wheteher it works as expected
-								   if not, GETOPT_GROKS_OPTIONAL_ARGS is undefined. @@@ */
-#else
-static char optstring[] = "+a:b:cC:d:f:hH:il:nm:rs:v";	/* it is possible that on some systems :: works but not +:. If so, please tell me! */
-#endif
-
-#ifdef HAVE_GETOPT_LONG
-static struct option longopts[] = {
-  {"always-readline", optional_argument, NULL, 'a'},
-  {"break-chars", required_argument, NULL, 'b'},
-  {"complete-filenames", no_argument, NULL, 'c'},
-  {"command-name", required_argument, NULL, 'C'},
-  {"debug", optional_argument, NULL, 'd'},
-  {"file", required_argument, NULL, 'f'},
-  {"help", no_argument, NULL, 'h'},
-  {"history-filename", required_argument, NULL, 'H'},
-  {"case-insensitive", no_argument, NULL, 'i'},
-  {"logfile", required_argument, NULL, 'l'},
-  {"no-warnings", no_argument, NULL, 'n'},
-  {"multiline-separator", optional_argument, NULL, 'm'},
-  {"remember", no_argument, NULL, 'r'},
-  {"version", no_argument, NULL, 'v'},
-  {"histsize", required_argument, NULL, 's'},
-  {0, 0, 0, 0}
-};
-#endif
+static int  output_queue_is_nonempty(void);
+static int  last_option_didnt_have_optional_argument = FALSE;
+static int  last_opt = -1;
+static int  bleach_input = FALSE;
 
 
 int
 main(int argc, char **argv)
 {
   char *command_name;
-
   init_completer();
   command_name = read_options_and_command_name(argc, argv);
+  /* by now, optind points to <command>, and &argv[optind] is <command>'s argv */
+  if (!isatty(STDIN_FILENO) && execvp(argv[optind], &argv[optind]) < 0)
+       /* if stdin is not a tty, just execute <command> */ 
+       myerror("Cannot execute %s", argv[optind]);	
   init_rlwrap();
   fork_child(command_name, argv);
   main_loop();
-  return 42;			/* not reached, but keeps the compiler happy... */
+  return 42;			/* not reached, but some compilers are unhappy without this ... */
 }
 
 
@@ -108,9 +90,13 @@ fork_child(char *command_name, char **argv)
     child_pid = pid;
   else {			/* child: */
     DPRINTF1(DEBUG_TERMIO, "preparing to execute %s", argv[optind]);
+    close_open_files_without_writing_buffers();
+   
     if (execvp(argv[optind], &argv[optind]) < 0) {
-      myerror("Cannot execute %s", argv[optind]);	
-      exit(EXIT_FAILURE);	/* stillborn child, parent will live on */
+      if (last_opt > 0 && last_option_didnt_have_optional_argument) /* e.g. 'rlwrap -a Password: sqlpus' will try to exec 'Password:' */
+	fprintf(stderr, "Did you mean '%s' to be an option argument?\nThen you should write '-%c%s', without the space(s)\n",
+		      argv[optind], last_opt, argv[optind]); 
+      myerror("Cannot execute %s", argv[optind]);   	/* stillborn child, parent will live on and display child's last gasps */
     }
   }
 }
@@ -199,6 +185,9 @@ main_loop()
 	}
 	buf[nread] = '\0';
 
+        if (bleach_input)
+	  bleach(buf);
+	
 	if (within_line_edit)	/* client output arrives while we're editing keyboard input  */
 	  save_rl_state(&saved_rl_state);
 
@@ -248,9 +237,10 @@ main_loop()
 		 mangle_string_for_debug_log(saved_rl_state.prompt, 40),
 		 strlen(saved_rl_state.prompt));
 	saved_rl_state.prompt[BUFFSIZE - 1] = '\0';	/* just to make sure.... */
+        
 	if (within_line_edit)
 	  restore_rl_state(&saved_rl_state);
-
+	
       }
 
       /* ----------------------------- key pressed: read stdin -------------------------*/
@@ -306,14 +296,13 @@ init_rlwrap()
   char *homedir, *histdir, *homedir_prefix;
 
   /* open debug file if necessary */
-#ifdef DEBUG
-  if (debug) {
+
+  if (debug) {    
     debug_fp = fopen(DEBUG_FILENAME, "w");
     if (!debug_fp)
       myerror("Couldn't open debug file %s", DEBUG_FILENAME);
     setbuf(debug_fp, NULL);
   }
-#endif
 
   DPRINTF0(DEBUG_TERMIO, "Initialising");
   init_terminal();
@@ -321,8 +310,9 @@ init_rlwrap()
 
   /* Determine rlwrap home dir and prefix for default history and completion filenames */
   homedir = (getenv("RLWRAP_HOME") ? getenv("RLWRAP_HOME") : getenv("HOME"));
-  homedir_prefix = (getenv("RLWRAP_HOME") ? add2strings(getenv("RLWRAP_HOME"), "/") :	/* use $RLWRAP_HOME/<command>_history */
-		    add2strings(getenv("HOME"), "/."));	/* use ~/.<command>_history */
+  homedir_prefix = (getenv("RLWRAP_HOME") ?                    /* is RLWRAP_HOME set?                */
+		    add2strings(getenv("RLWRAP_HOME"), "/") :  /* use $RLWRAP_HOME/<command>_history */
+		    add2strings(getenv("HOME"), "/."));	       /* if not, use ~/.<command>_history   */
 
   /* Determine history file name and check its existence and permissions */
 
@@ -353,12 +343,6 @@ init_rlwrap()
   stifle_history(histsize);
   read_history(history_filename);	/* ignore errors here: history file may not yet exist, but will be created on exit */
 
-  /* Get last input line from history and put it in previous_line
-     The last added history_entry is always at history_length + history_base - 1 */
-  if (history_length)
-    mystrlcpy(previous_line,
-	      history_get(history_length + history_base - 1)->line, BUFFSIZE);
-
 
   /* Determine completion file name (completion files are never written to,
      and ignored when unreadable or non-existent) */
@@ -382,21 +366,83 @@ init_rlwrap()
 
 /*
  * On systems where getopt doens't handle optional argments, warn the user whenever an
- * argument of the form -<letter> is seen
+ * argument of the form -<letter> is seen, or whenever the argument is the last item on the command line
+ * (e.g. 'rlwrap -a command', which will be parsed as 'rlwrap --always-readline=command')
  */
+
 static char *
-check_optarg(char opt)
+check_optarg(char opt, int remaining)
 {
+  if (!optarg)
+    last_option_didnt_have_optional_argument = TRUE; /* if command is not found, suggest that it may have been meant
+						        as optional argument (e.g. 'rlwrap -a password sqlplus' will try to
+						        execute 'password' ) */
 #ifndef GETOPT_GROKS_OPTIONAL_ARGS
-  if (optarg && optarg[0] == '-' && isalpha(optarg[1]))
-    mywarn
-      ("on this system, %s doesn't have optional arguments, so '%s' is taken as an argument to the -%c option",
-       program_name, optarg, opt);
+  if (optarg &&    /* is there an optional arg? have a look at it: */
+      ((optarg[0] == '-' && isalpha(optarg[1])) || /* looks like next option */
+       remaining == 0)) /* or is last item on command line */
+
+  mywarn
+      ("on this system, the getopt() library function doesn't\n"
+       "grok optional arguments, so '%s' is taken as an argument to the -%c option\n"
+       "Is this what you meant? If not, please provide an argument"
+#ifdef HAVE_GETOPT_LONG
+       ", or use a --long-option.\n",
+#else
+       ".\n",
 #endif
+       optarg, opt);
+#endif
+  
   return optarg;
 }
 
 
+#ifdef GETOPT_GROKS_OPTIONAL_ARGS
+static char optstring[] = "+:a::b:cC:d::D:f:F:hH:il:nm::P:rs:tv";	/* +: is not really documented. configure checks wheteher it works as expected
+					        			   if not, GETOPT_GROKS_OPTIONAL_ARGS is undefined. @@@ */
+#else
+static char optstring[] = "+:a:b:cC:d:D:f:F:hH:il:nm:P:rs:tv";	/* it is possible that on some systems :: works but not +:. If so, please tell me! */
+#endif
+
+#ifdef HAVE_GETOPT_LONG
+static struct option longopts[] = {
+  {"always-readline", 		optional_argument, 	NULL, 'a'},
+  {"break-chars", 		required_argument, 	NULL, 'b'},
+  /*  {"bleach",                    no_argument,		NULL, 'B'}, still not usable */
+  {"complete-filenames", 	no_argument, 		NULL, 'c'},
+  {"command-name", 		required_argument, 	NULL, 'C'},
+  {"debug", 			optional_argument, 	NULL, 'd'},
+  {"history-no-dupes", 		required_argument, 	NULL, 'D'},
+  {"file", 			required_argument, 	NULL, 'f'},
+  {"history-format", 		required_argument, 	NULL, 'F'},
+  {"help", 			no_argument, 		NULL, 'h'},
+  {"history-filename", 		required_argument, 	NULL, 'H'},
+  {"case-insensitive", 		no_argument, 		NULL, 'i'},
+  {"logfile", 			required_argument, 	NULL, 'l'},
+  {"no-warnings", 		no_argument, 		NULL, 'n'},
+  {"multi-line", 		optional_argument, 	NULL, 'm'},
+  {"pre-given", 		required_argument, 	NULL, 'P'},
+  {"remember", 			no_argument, 		NULL, 'r'},
+  {"version", 			no_argument, 		NULL, 'v'},
+  {"histsize", 			required_argument, 	NULL, 's'},
+  {"test-terminal",  		no_argument, 		NULL, 't'},
+  {0, 0, 0, 0}
+};
+#endif
+
+
+static const char *current_option(int opt, int longindex) {
+  static char buf[BUFFSIZE];
+#ifdef HAVE_GETOPT_LONG
+  if (longindex >=0) {
+    sprintf(buf, "--%s", longopts[longindex].name);
+    return buf;
+  }	
+#endif
+  sprintf(buf, "-%c", opt);
+  return buf;
+}
 
 char *
 read_options_and_command_name(int argc, char **argv)
@@ -405,7 +451,9 @@ read_options_and_command_name(int argc, char **argv)
   char *opt_C = NULL;
   int opt_b = FALSE;
   int opt_f = FALSE;
-
+  int remaining = -1; /* remaining number of argumants on command line */
+  int longindex = -1; /* index of current option in longopts[], set by getopt_long */
+  
   full_program_name = mysavestring(argv[0]);
   program_name = mysavestring(mybasename(full_program_name));	/* normally "rlwrap"; needed by myerror() */
   rl_basic_word_break_characters = " \t\n\r(){}[],+-=&^%$#@\";|\\";
@@ -414,33 +462,38 @@ read_options_and_command_name(int argc, char **argv)
 
   while (1) {
 #ifdef HAVE_GETOPT_LONG
-    c = getopt_long(argc, argv, optstring, longopts, NULL);
+    c = getopt_long(argc, argv, optstring, longopts, &longindex);
 #else
     c = getopt(argc, argv, optstring);
 #endif
 
     if (c == EOF)
       break;
-
+    last_option_didnt_have_optional_argument = FALSE;
+    remaining = argc - optind;
+    last_opt = c;    
     switch (c) {
       case 'a':
 	always_readline = TRUE;
-	if (check_optarg('a'))
+	if (check_optarg('a', remaining))
 	  password_prompt_search_string = mysavestring(optarg);
 	break;
       case 'b':
 	rl_basic_word_break_characters = add3strings("\r\n \t", optarg, "");
 	opt_b = TRUE;
 	break;
+      case 'B':
+	bleach_input = TRUE;
+	break;
       case 'c':
 	complete_filenames = TRUE;
 	break;
       case 'C':
-	opt_C = mysavestring(optarg);	/* @@@ optarg is a GNU extension */
+	opt_C = mysavestring(optarg);	
 	break;
       case 'd':
 #ifdef DEBUG
-	if (check_optarg('d'))
+	if (check_optarg('d', remaining))
 	  debug = atoi(optarg);
 	else
 	  debug = DEBUG_ALL;
@@ -448,16 +501,28 @@ read_options_and_command_name(int argc, char **argv)
 	mywarn
 	  ("To use debugging, configure %s with --enable-debug and rebuild",
 	   program_name);
-	debug = 1;
 #endif
 	break;
+
+      case 'D': 
+        history_duplicate_avoidance_policy=atoi(optarg);
+	if (history_duplicate_avoidance_policy < 0 || history_duplicate_avoidance_policy > 2)
+	  myerror("%s option with illegal value %d, should be 0, 1 or 2",
+		  current_option('D', longindex), history_duplicate_avoidance_policy);
+	break;
       case 'f':
 	feed_file_into_completion_list(optarg);
 	opt_f = TRUE;
 	break;
+      case 'F':
+        history_format = mysavestring(optarg);
+	if (isspace(history_format[0]))
+	  myerror("%s option argument should start with non-space", current_option('F', longindex));
+	if(history_format[0] == '%' && history_format[1] != ' ')
+	  myerror("if %s option argument starts with '%%', is should start with '%% ' i.e. '%%<space>'", current_option('F', longindex));
+	break;
       case 'h':
 	usage(EXIT_SUCCESS);		/* will call exit() */
-
       case 'H':
 	history_filename = mysavestring(optarg);
 	break;
@@ -477,7 +542,12 @@ read_options_and_command_name(int argc, char **argv)
 	mywarn("the -m option doesn't work on this system");
 #endif
 	multiline_separator =
-	  (check_optarg('m') ? mysavestring(optarg) : " \\ ");
+	  (check_optarg('m', remaining) ? mysavestring(optarg) : " \\ ");
+	break;
+      case 'P':
+        pre_given = mysavestring(optarg);
+	always_readline = TRUE; /* pre_given does not work well with transparent mode */
+	one_shot_rlwrap = TRUE;
 	break;
       case 'r':
 	remember_for_completion = TRUE;
@@ -489,16 +559,23 @@ read_options_and_command_name(int argc, char **argv)
 	  histsize = -histsize;
 	}
 	break;
+#ifdef DEBUG
+      case 't':
+	test_terminal();
+        exit(EXIT_SUCCESS);
+#endif
       case 'v':
-	printf("%s %s\n", program_name, VERSION);
-	exit(1);
+	printf("rlwrap %s\n",  VERSION);
+	exit(EXIT_SUCCESS);
       case '?':
-	myerror("unknown option -- %c\ntry '%s --help' for more information",
-		optopt, full_program_name);
+	assert(optind > 0);
+	myerror("unrecognised option %s\ntry '%s --help' for more information",
+		argv[optind-1], full_program_name);
       case ':':
+	assert(optind > 0);
 	myerror
-	  ("option requires an argument -- %c\ntry '%s --help' for more information",
-	   optopt, full_program_name);
+	  ("option %s requires an argument \ntry '%s --help' for more information",
+	   argv[optind-1], full_program_name);
       default:
 	usage(EXIT_FAILURE);
     }
@@ -509,7 +586,8 @@ read_options_and_command_name(int argc, char **argv)
       add3strings(rl_basic_word_break_characters, "/.", "");
   }
 
-  if (optind >= argc)
+  
+  if (optind >= argc) /* rlwrap -a -b -c with no command specified */
     usage(EXIT_FAILURE);
 
   if (opt_C) {
@@ -553,7 +631,7 @@ read_options_and_command_name(int argc, char **argv)
  * asynchronously, keeping a queue of pending output. The readline
  * line handler calls put_in_output_queue(usder_input) , while
  * main_loop calls flush_output_queue() as long as there is something
- * in the queue
+ * in the queue.
  */
 
 static char *output_queue; /* NULL when empty */
@@ -568,12 +646,12 @@ void
 put_in_output_queue(char *stuff)
 {
   if (!output_queue)
-    output_queue = mysavestring(stuff);
+    output_queue = mysavestring(stuff);           /* allocate queue (no need to free anything) */
   else {
     char *old_queue = output_queue;
 
-    output_queue = add2strings(old_queue, stuff);
-    free(old_queue);
+    output_queue = add2strings(old_queue, stuff); /* allocate new queue */
+    free(old_queue);                              /* free the old one   */
   }
 }
 
@@ -586,10 +664,10 @@ flush_output_queue()
   if (!output_queue)
     return;
   queuelen = strlen(output_queue);
-  how_much = min(BUFFSIZE, queuelen);
+  how_much = min(BUFFSIZE, queuelen); /* never write more than BUFFSIZE in one go @@@ is this never too much? */
   nwritten = write(master_pty_fd, output_queue, how_much);
   if (nwritten < 0) {
-    struct timeval ten_millisecs = { 0, 100000 };
+    struct timeval ten_millisecs = { 0, 10000 };
     switch (nwritten) {
       case EINTR:
 	return;
@@ -617,11 +695,24 @@ void
 cleanup_rlwrap_and_exit(int status)
 {
   DPRINTF0(DEBUG_TERMIO, "Cleaning up");
-  if (write_histfile && history_total_bytes() > 0)	/* avoid creating files like .nsloolup_history after typo */
+  if (write_histfile && (histsize==0 ||  history_total_bytes() > 0))
+    /* avoid creating .speling_error_history after typo */
     write_history(history_filename);	/* ignore errors */
   close_logfile();
   if (terminal_settings_saved && (tcsetattr(STDIN_FILENO, TCSAFLUSH, &saved_terminal_settings) < 0))	/* reset terminal */
-    ;				/*     mywarn ("tcsetattr error on stdin"); */
-                                /* NOT myerror, as this would call cleanup_rlwrap_and_exit, causing infinite recursion */
-  exit(status != EXIT_SUCCESS ? status : WEXITSTATUS(childs_exit_status));	/* signal failure, or else propagate child's exit status */
+    /* ignore (almost dead anyway) */ ;	 /* mywarn ("tcsetattr error on stdin"); */
+                                         /* NOT myerror, as this would call cleanup_rlwrap_and_exit again*/
+
+#ifdef WTERMSIG                 /* is there any system without this macro? */
+  if (status != EXIT_SUCCESS)
+    exit(status);               /* rlwrap itself has failed, rather than the wrapped command */
+  else if (WIFSIGNALED(childs_exit_status)) 
+    suicide_by(WTERMSIG(childs_exit_status)); /* child terminated by signal, make rlwrap's
+						 parent believe rlwrap was killed by it */ 
+  else
+    exit(WEXITSTATUS(childs_exit_status)); /* propagate child's exit status */
+#else
+  exit(status != EXIT_SUCCESS ? status : WEXITSTATUS(childs_exit_status)); /* WEXITSTATUS has a default definition in rlwrap.h */
+#endif
+  
 }
