@@ -29,6 +29,7 @@
 #endif
 
 
+
 #include <sys/stat.h>
 #include <sys/ioctl.h>
 #include <fcntl.h>
@@ -46,7 +47,7 @@
 #include <stdlib.h>
 
 
-#include <assert.h>
+
 
 #if TIME_WITH_SYS_TIME
 # include <sys/time.h>
@@ -74,7 +75,8 @@
 #  ifndef HAVE_STRRCHR
 #    define strrchr rindex
 #  endif
-   char *strchr (), *strrchr ();
+char *strchr(), *strrchr();
+
 #  ifndef HAVE_MEMMOVE
 #    define memmove(d, s, n) bcopy ((s), (d), (n))
 #  endif
@@ -94,14 +96,23 @@
 
 
 
+
 #define BUFFSIZE 512
 
+#ifndef MAXPATHLEN
+#define MAXPATHLEN 512
+#endif
 
-#ifdef  HAVE_SNPRINTF  /* don't rely on the compiler understanding variadic macros */
-# define snprintf1(buf,bufsize,format,arg1) snprintf(buf,bufsize,format,arg1)
+
+#ifdef  HAVE_SNPRINTF		/* don't rely on the compiler understanding variadic macros */
+# define snprintf0(buf,bufsize,format)    		snprintf(buf,bufsize,format)
+# define snprintf1(buf,bufsize,format,arg1) 		snprintf(buf,bufsize,format,arg1)
+# define snprintf2(buf,bufsize,format,arg1,arg2) 	snprintf(buf,bufsize,format,arg1,arg2)
 #else
-# define snprintf1(buf,bufsize,format,arg1) sprintf(buf,format,arg1)
-# define vsnprintf(buf,bufsize,format,ap)   vsprintf(buf,format,ap)
+# define snprintf0(buf,bufsize,format)      		sprintf(buf,format)
+# define snprintf1(buf,bufsize,format,arg1) 		sprintf(buf,format,arg1)
+# define snprintf2(buf,bufsize,format,arg1,arg2) 	sprintf(buf,format,arg1,arg2)
+# define vsnprintf(buf,bufsize,format,ap)   		vsprintf(buf,format,ap)
 #endif
 
 
@@ -116,19 +127,26 @@
 
 #ifdef SPY_ON_READLINE
 # ifndef HOMEGROWN_REDISPLAY
+
 #  define MAYBE_MULTILINE 1
 # endif
 #endif
 
 #ifdef MAYBE_MULTILINE
-extern int _rl_horizontal_scroll_mode; /* Spying on readline's private life .... */
+extern int _rl_horizontal_scroll_mode;	/* Spying on readline's private life .... */
+
 #  define redisplay_multiple_lines (!_rl_horizontal_scroll_mode)
 #else
 #  define redisplay_multiple_lines FALSE
 #endif
 
+
+
+
+
+
 /* in main.c: */
-extern int master_pty_fd; 
+extern int master_pty_fd;
 extern int slave_pty_fd;
 extern FILE *debug_fp;
 extern char *program_name, *command_name;
@@ -139,22 +157,33 @@ extern int i_am_child;
 extern int nowarn;
 extern int debug;
 extern char *password_prompt_search_string;
+extern char *history_format;
+extern int ignore_queued_input;
+extern int history_duplicate_avoidance_policy;
+extern int one_shot_rlwrap;
+
+/* now follow the possible values for history_duplicate_avoidance_policy: */
+#define KEEP_ALL_DOUBLES 		0
+#define ELIMINATE_SUCCESIVE_DOUBLES	1
+#define ELIMINATE_ALL_DOUBLES  		2
+
 
-void init_rlwrap(void);
 void cleanup_rlwrap_and_exit(int status);
-void main_loop(int pty_fd);
+void put_in_output_queue(char *stuff);
+
 
 
 /* in signals.c */
 extern int child_is_dead, childs_exit_status, sigterm_received;
 void install_signal_handlers(void);
-void block_signals(int* sigs);
-void unblock_signals(int* sigs);
+void block_signals(int *sigs);
+void unblock_signals(int *sigs);
 void ignore_sigchld(void);
+void suicide_by(int sig);
 
 /* in utils.c */
-void usage(void);
-void mywarn(const char  *message, ...);
+void usage(int status);
+void mywarn(const char *message, ...);
 void myerror(const char *message, ...);
 void *mymalloc(size_t size);
 char *mybasename(char *filename);
@@ -162,53 +191,72 @@ char *mydirname(char *filename);
 char *mystrndup(const char *string, int len);
 void mystrlcpy(char *dst, const char *src, size_t size);
 void mystrlcat(char *dst, const char *src, size_t size);
-char *mysavestring(const char*string);
+char* mystrstr(const char *haystack, const char *needle);
+char *mysavestring(const char *string);
 char *add3strings(const char *str1, const char *str2, const char *str3);
+void close_open_files_without_writing_buffers(void);
+
 #define add2strings(a,b)  add3strings(a,b,"")
-char *mangle_string_for_debug_log(char *string, int maxlen);
+char *mangle_string_for_debug_log(const char *string, int maxlen);
+char *search_and_replace(char *patt, char *repl, const char *string,
+			 int cursorpos, int *line, int *col);
+char *first_of(char **strings);
+char *as_string(int i);
+size_t filesize(const char *filename);
 void open_logfile(const char *filename);
 void write_logfile(const char *str);
 void close_logfile(void);
+void change_working_directory(void);
+extern char slaves_working_directory[];
+void bleach(char *buffer);
+char *append_and_expand_history_format(char *line);
 
 /* in readline.c: */
-extern struct rl_state {      /* struct to save readline text buffer and point */
+extern struct rl_state
+{				/* struct to save readline text buffer and point */
   char text[BUFFSIZE];
   char prompt[BUFFSIZE];
-  int  point;
-  int  already_saved;         /* flag set when saved, cleared when restored */ 
+  int point;
+  int already_saved;		/* flag set when saved, cleared when restored */
 } saved_rl_state;
 
-void save_rl_state (struct rl_state *);
-void restore_rl_state (struct rl_state *);
+void save_rl_state(struct rl_state *);
+void restore_rl_state(struct rl_state *);
 void init_readline(char *);
 void line_handler(char *);
 void my_redisplay(void);
 
-extern  int within_line_edit, transparent;
-extern  char previous_line[];
+extern int within_line_edit, transparent;
+extern char *multiline_separator;
+extern char *pre_given;
 
 /* in pty.c: */
-pid_t my_pty_fork ( int *, const struct termios *, const struct winsize *);
+pid_t my_pty_fork(int *, const struct termios *, const struct winsize *);
 int slave_is_in_raw_mode(void);
+struct termios *get_pterm_slave(void);
 void mirror_slaves_echo_mode(void);
+void completely_mirror_slaves_terminal_settings(void);
+void write_EOF_to_master_pty(void);
+void write_EOL_to_master_pty(char *);
 
 /* in ptytty.c */
-int  ptytty_get_pty (int *fd_tty, const char **ttydev);
-int  ptytty_get_tty (const char *ttydev);
-int  ptytty_control_tty (int fd_tty, const char *ttydev);
-int  ptytty_openpty(int  *amaster,  int *aslave, const char **name);
+int ptytty_get_pty(int *fd_tty, const char **ttydev);
+int ptytty_get_tty(const char *ttydev);
+int ptytty_control_tty(int fd_tty, const char *ttydev);
+int ptytty_openpty(int *amaster, int *aslave, const char **name);
 
 
 
 /* in completion.c */
-void init_completer(const char *);
-void feed_into_completion_list(const char *);
+void init_completer(void);
+void feed_file_into_completion_list(const char *);
+void feed_line_into_completion_list(const char *);
 char *my_completion_function(char *, int);
 
-extern int  remember_for_completion;
+extern int remember_for_completion;
 
 /* in term.c */
-int redisplay;                         /* TRUE when user input should be visible */  
+int redisplay;			/* TRUE when user input should be visible */
 void init_terminal(void);
 void set_echo(int);
 void prepare_terminal(void);
@@ -218,12 +266,17 @@ void clear_line(void);
 void clear_the_screen(void);
 void curs_up(void);
 void curs_down(void);
-int  my_putchar(int c);
-void cursor_hpos (int col);
+void test_terminal(void);
+int my_putchar(int c);
+void cursor_hpos(int col);
 extern struct termios saved_terminal_settings;
 extern int terminal_settings_saved;
 extern struct winsize window_size;
-extern char *term_backspace, term_eof, term_stop, *term_cursor_hpos, *term_cursor_up, *term_cursor_down;
+extern char *term_backspace, term_eof, term_stop, *term_cursor_hpos,
+  *term_cursor_up, *term_cursor_down, *term_newline;
+
+/* in completion.rb */
+int completion_is_case_sensitive;
 
 /* some handy macros */
 #ifndef TRUE
@@ -253,12 +306,12 @@ extern char *term_backspace, term_eof, term_stop, *term_cursor_hpos, *term_curso
 #  define DEBUG_TERMIO                     	 1
 #  define DEBUG_SIGNALS                    	 2
 #  define DEBUG_READLINE                   	 4
-#  define DEBUG_ALL                        	 (DEBUG_TERMIO | DEBUG_SIGNALS | DEBUG_READLINE)  
+#  define DEBUG_ALL                        	 (DEBUG_TERMIO | DEBUG_SIGNALS | DEBUG_READLINE)
 
 #ifdef __GNUC__
-#  define WHERE                                	  fprintf(debug_fp, "%10s:%-3d\t%-25s ", __FILE__, __LINE__, __FUNCTION__)
+#  define WHERE                                	  fprintf(debug_fp, "%13s:%-3d\t%-25s ", __FILE__, __LINE__, __FUNCTION__)
 #else
-#  define WHERE                                   fprintf(debug_fp, "%10s:%-3d\t", __FILE__, __LINE__)
+#  define WHERE                                   fprintf(debug_fp, "%13s:%-3d\t", __FILE__, __LINE__)
 #endif
 
 #  define NLFLUSH                                 fputc('\n', debug_fp) ; fflush(debug_fp)
@@ -268,15 +321,12 @@ extern char *term_backspace, term_eof, term_stop, *term_cursor_hpos, *term_curso
 #  define DPRINTF3(mask, format,arg1, arg2, arg3) if (debug & mask) {WHERE; fprintf(debug_fp, format, arg1, arg2, arg3); NLFLUSH; }
 #  define ERRMSG(b)                               (b && (errno != 0) ? add3strings("(", strerror(errno), ")") : "" )
 #else
-#  define DPRINTF0(mask, format) 
-#  define DPRINTF1(mask, format,arg) 
+#  define DPRINTF0(mask, format)
+#  define DPRINTF1(mask, format,arg)
 #  define DPRINTF2(mask, format,arg1, arg2)
 #  define DPRINTF3(mask, format,arg1, arg2, arg3)
-#  define ERRMSG(b) 
-#  define NDEBUG    /* disable assertions */
+#  define ERRMSG(b)
+#  define NDEBUG		/* disable assertions */
 #endif
 
-
-
-
-
+#include <assert.h>
